""" Spike sorting using Bayes Optimal Template Matching (BOTM).
Needs spike trains from a previous analysis (e.g. spike detection).

First, a spike detection is run until "Spikes for clustering" spikes
are found (or no more data is available for detection because 
"Initialization segments" or all selected segments have been used).
The spikes are then clustered using a Gaussian Mixture Model, cluster
counts from "Minimum clusters" to "Maximum clusters" are tried.
Data strips of "Spike samples" length are considered, their minimum
is on "Align sample". The feature used for alignment is given by
"Align on".

After clustering, clusters containing less than "Minimum cluster size"
spikes are discarded and clusters are merged when the euclidean distance
of their means is smaller than "Maximum merge distance".

After "Initialization Segments" or after all selected segments have been
processed, the real sorting starts again at the first segment. No output
is generated before this jump! If less than "Spikes for clustering"
spikes have been found before this point, no sorting will be created.

New units and their corresponding spike trains are created. If
"Create template spikes" is checked, a spike with the prototypical 
waveform used for sorting is created in each segment for each unit.

If there are artifact epochs that should not be considered for spike
detection, check "Ignore artifact epochs" and set "Artifact Tag" to 
the value of their 'tag' annotation.
"""
from spykeutils.plugin import analysis_plugin, gui_data
from spykeutils import SpykeException
import copy
import scipy as sp
import quantities as pq

import neo
from botmpy.nodes import ABOTMNode
from botmpy.common import TimeSeriesCovE, dict_list_to_ndarray, merge_epochs

tag_prop = gui_data.ValueProp(False)

class SpikeSortingPlugin(analysis_plugin.AnalysisPlugin):
    """ Spike Sorting BOTM """
    
    tf = gui_data.IntItem('Spike samples', default=49)
    align_sample = gui_data.IntItem('Align sample', default=14)
    alignment_option = gui_data.ChoiceItem(
        'Align on', ('Minimum', 'Maximum', 'Energy', 'None'))
    min_clusters = gui_data.IntItem('Minmum clusters', default=1, min=1)
    max_clusters = gui_data.IntItem('Maximum clusters', default=14, min=1)
    init_segments = gui_data.IntItem('Initialization segments', default=10)
    cluster_spikes = gui_data.IntItem('Spikes for clustering', default=4000)
    min_cluster = gui_data.IntItem('Minimum cluster size', default=50)
    pca_features = gui_data.IntItem('Principal components for clustering', 
                                    default=10, min=2)
    amplitude_features = gui_data.BoolItem('Use amplitude features for clustering',
                                           default=True)
    merge_cluster = gui_data.FloatItem('Maximum merge distance', default=0.0)
    template_spike = gui_data.BoolItem('Create template spikes', default=True)
    ignore_artifacts = gui_data.BoolItem(
        'Remove artifact epochs').set_prop('display', store=tag_prop)
    artifact_tag = gui_data.StringItem(
        'Artifact tag').set_prop('display', active=tag_prop)
    

    def __init__(self):
        super(SpikeSortingPlugin, self).__init__()
    
    def get_name(self):
        return 'Spike Sorting'

    def start(self, current, selections):
        if len(current.recording_channel_groups()) != 1:
            raise SpykeException(
                'Please select exactly one recording channel group.')
        trains = current.spike_trains_by_segment()
        
        if not current.units() or not trains:
            raise SpykeException(
                'Please select at least on unit with spike trains for clustering.')
        
        rcg = current.recording_channel_groups()[0]
        new_units = {}
        
        current.progress.begin('Sorting spikes')
        signals = current.analog_signals_by_segment(3)
        trains = current.spike_trains_by_segment()
        epochs = current.epochs()
        current.progress.set_ticks(len(signals) + min(len(signals), self.init_segments))
        
        if not signals:
            raise SpykeException('No analog signals found.')
            
        num_channels = len(signals.values()[0])
        
        self.ce = TimeSeriesCovE(tf_max=self.tf, nc=num_channels)

        align_kinds = ['min', 'max', 'energy', 'none']
        self.sorter = ABOTMNode(
            tf=self.tf,
            ce=self.ce,
            ovlp_taus=None,
            chunk_size=100000,
            verbose=15,
            learn_noise='det',
            det_cls=None,
            det_kwargs=None,
            det_limit=self.cluster_spikes,
            learn_templates=self.align_sample,
            learn_templates_rsf=1.0,
            spk_pr_bias=(1000, self.tf / 2),
            det_forget=0,
            align_kind=align_kinds[self.alignment_option],
            clus_algo='gmm',
            clus_params={'min_clusters': self.min_clusters,
                         'max_clusters': self.max_clusters},
            clus_min_size=self.min_cluster,
            clus_merge_dist=self.merge_cluster,
            clus_pca_features=self.pca_features,
            clus_use_amplitudes=self.amplitude_features)

        i_seg = 0
        for init_done in range(2):  # Two times: initialization and sorting run
            for seg, sigs in signals.iteritems():  # For each segment...
                # Create data for spike sorting
                sig, eps = self.prepare_data(seg, sigs, epochs)
                
                # Merge all selected spike trains for this segment
                seg_trains = []
                srate = sigs[0].sampling_rate
                for st in trains[seg]:
                    seg_trains.append(sp.asarray((st * srate).simplified))
                if not seg_trains:
                    raise SpykeException(
                        'No previous spikes in segment %s' % (seg.name or '???'))
                spike_array = sp.concatenate(seg_trains)
                
                # Perform sorting
                sorting = []
                for ep in eps:
                    self.sorter.reset_history()
                    
                    data = sig[ep[0]:ep[1], :]
                    ep_spikes = spike_array[
                        (spike_array > ep[0]) & (spike_array < ep[1])]
                    self.sorter(data, ep_spikes)

                    sorting.append((copy.deepcopy(self.sorter.rval), ep[0]))
                    
                if init_done:
                    # Merge sortings
                    out = {}
                    for ss, offset in sorting:
                        for k in ss:
                            if k not in out:
                                out[k] = []
                            out[k].extend(
                                ss[k] + offset + self.align_sample - self.tf / 2)
                                
                    # Create Neo objects
                    for uid, spikes in out.iteritems():
                        if uid not in new_units:
                            new_units[uid] = neo.Unit('Unit %d' % (uid + 1))
                            new_units[uid].recordingchannelgroup = rcg
                        
                        # Create and attach spike train
                        st = neo.SpikeTrain(
                            spikes * sigs[0].sampling_period.simplified + sigs[0].t_start,
                            t_start=sigs[0].t_start, t_stop=sigs[0].t_stop)
                            
                        new_units[uid].spiketrains.append(st)
                        st.unit = new_units[uid]
                        seg.spiketrains.append(st)
                        st.segment = seg
                        
                        # Create and attach template spike
                        spike = neo.Spike(
                            0*pq.s, 
                            waveform=self.sorter.bank[uid].xi * sigs[0].units,
                            sampling_rate=sigs[0].sampling_rate,
                            left_sweep=(self.align_sample / sigs[0].sampling_rate).simplified)
                        
                        new_units[uid].spikes.append(spike)
                        spike.unit = new_units[uid]
                        seg.spikes.append(spike)
                        spike.segment = seg
                    
                current.progress.step()
                
                if not init_done:
                    i_seg += 1
                    if i_seg >= self.init_segments:
                        break
            
            if self.sorter._cluster == self.sorter._cluster_init:  
                self.sorter._cluster()  # No clustering yet, do it now

        # Attach created units        
        rcg.units.extend(new_units.values())
        
        current.progress.done()
        current.refresh_view()
                
                
    def prepare_data(self, seg, sigs, epochs):
        try:
            sig = sp.vstack([sp.asarray(s) for s in sigs]).T
        except ValueError:
            segment_desc = ''
            if seg.name:
                segment_desc = 'in segment %s ' % seg.name
            raise SpykeException(
                'Error: Analog signals ' + segment_desc + 
                'do not have the same length!')
                
        # Cut artifacts
        eps = []  # List of index tuples for non-artifact epochs
        arts = []  # List of artifact epochs
        if self.ignore_artifacts and seg in epochs:
            arts = [e for e in epochs[seg] 
                    if e.annotations.get('tag', None) == self.artifact_tag]
        
        sampling_rate = sigs[0].sampling_rate
        last_idx = 0
        for a in arts:
            eps.append((last_idx,
                        int((a.time * sampling_rate).simplified)))
            last_idx = int(
                ((a.time + a.duration) * sampling_rate).simplified)
        eps.append((last_idx, sig.shape[0]))
        eps = merge_epochs(eps, min_dist=self.tf)
            
        return sig, eps
        
    def configure(self):
        from PyQt4.QtGui import QMessageBox
        
        super(SpikeSortingPlugin, self).configure()
        while self.tf % 2 == 0:
            QMessageBox.warning(None, 'Unable to set parameters',
                '"Spike samples" needs to be uneven!')
            super(SpikeSortingPlugin, self).configure()